package com.tzld.longarticle.recommend.server.service.recommend.filter;

import com.tzld.longarticle.recommend.server.common.ContentCountMonitor;
import com.tzld.longarticle.recommend.server.common.CostMonitor;
import com.tzld.longarticle.recommend.server.common.ThreadPoolFactory;
import com.tzld.longarticle.recommend.server.common.enums.recommend.RankStrategyEnum;
import com.tzld.longarticle.recommend.server.mapper.longArticle.LongArticleBaseMapper;
import com.tzld.longarticle.recommend.server.model.dto.Content;
import com.tzld.longarticle.recommend.server.service.ServiceBeanFactory;
import com.tzld.longarticle.recommend.server.service.recommend.filter.strategy.*;
import com.tzld.longarticle.recommend.server.util.CommonCollectionUtils;
import com.tzld.longarticle.recommend.server.util.DateUtils;
import com.tzld.longarticle.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

import static com.tzld.longarticle.recommend.server.common.constant.SceneConstants.FWH_COLD_START;

@Slf4j
@Service
public class FilterService {
    private final ExecutorService pool = ThreadPoolFactory.filterPool();

    @Autowired
    LongArticleBaseMapper longArticleBaseMapper;
    @Autowired
    private RedisTemplate<String, String> redisTemplate;


    public FilterResult filter(FilterParam param, Boolean preFilter) {
        return contentFilter(param, preFilter);
    }

    private FilterResult contentFilter(FilterParam param, Boolean preFilter) {
        long start = System.currentTimeMillis();
        FilterResult result = new FilterResult();

        List<FilterStrategy> strategies = getAllStrategies(param);

        // 已同步待发布内容 预过滤
        if (!preFilter) {
            String dateStr = DateUtils.getCurrentDateStr("yyyyMMdd");
            String redisKey = "ContentPreFilterJob:" + dateStr + ":" + param.getPlanId() + "-" + param.getAccountId();
            String value = redisTemplate.opsForValue().get(redisKey);
            if (StringUtils.hasText(value)) {
                Long minTime = longArticleBaseMapper.getMinGzhWaitingPublishContent(param.getPlanId(), param.getAccountId());
                Long todayMinTime = DateUtils.getTodayStart();
                if (Objects.nonNull(minTime) && minTime < todayMinTime) {
                    strategies = getUnPreFilterStrategies(param);
                }
            }
        } else {
            strategies = getPreFilterStrategies(param);
        }
        log.info("FilterStrategy {}", JSONUtils.toJson(CommonCollectionUtils.toList(strategies,
                s -> s.getClass().getSimpleName())));
        CountDownLatch cdl = new CountDownLatch(strategies.size());
        List<Future<FilterResult>> futures = new ArrayList<>();
        for (final FilterStrategy strategy : strategies) {
            Future<FilterResult> future = pool.submit(() -> {
                try {
                    long t1 = System.currentTimeMillis();
                    FilterResult ret = strategy.filter(param);
                    long t2 = System.currentTimeMillis();
                    CostMonitor.logCost("Filter", strategy.getClass().getSimpleName(), t2 - t1);
                    ContentCountMonitor.logCount("Filter", strategy.getClass().getSimpleName(),
                            Objects.isNull(ret.getContentIds()) ? 0 : ret.getContentIds().size());
                    return ret;
                } finally {
                    cdl.countDown();
                }
            });
            futures.add(future);
        }
        try {
            cdl.await();
        } catch (InterruptedException e) {
            log.error("filter error", e);
            return null;
        }

        List<List<String>> contentIdsList = new ArrayList<>();
        List<Content> filterContents = new ArrayList<>();
        Set<String> filterContentIds = new HashSet<>();
        for (Future<FilterResult> f : futures) {
            try {
                if (Objects.isNull(f)) {
                    continue;
                }
                FilterResult filterResult = f.get();
                contentIdsList.add(filterResult.getContentIds());
                if (CollectionUtils.isNotEmpty(filterResult.getFilterContent())) {
                    for (Content content : filterResult.getFilterContent()) {
                        if (!filterContentIds.contains(content.getId())) {
                            filterContentIds.add(content.getId());
                            filterContents.add(content);
                        }
                    }
                }
            } catch (Exception e) {
                log.error("future get error ", e);
            }
        }
        if (CollectionUtils.isEmpty(contentIdsList)) {
            return result;
        }
        List<String> contentIds = param.getContents().stream().map(Content::getId).collect(Collectors.toList());
        for (int i = 0; i < contentIdsList.size(); ++i) {
            contentIds.retainAll(contentIdsList.get(i));
        }
        result.setContentIds(contentIds);
        result.setFilterContent(filterContents);
        log.info("filter cost:{}", System.currentTimeMillis() - start);
        return result;
    }

    private List<FilterStrategy> getAllStrategies(FilterParam param) {
        List<FilterStrategy> strategies = new ArrayList<>();
        strategies.add(ServiceBeanFactory.getBean(BadStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(SensitiveStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(DeDuplicationStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(KeywordStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(ArticlePromotionStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(TodayPublishStrategy.class));
//        strategies.add(ServiceBeanFactory.getBean(VideoPoolBadAuditStrategy.class));
        if (param.getScene().equals(FWH_COLD_START)) {
            strategies.add(ServiceBeanFactory.getBean(HistoryTitleForFwhColdStartStrategy.class));
        } else {
            if (Objects.equals(param.getStrategy(), RankStrategyEnum.INFINITE_STRATEGY.getStrategy())) {
                strategies.add(ServiceBeanFactory.getBean(InfiniteHisTitleStrategy.class));
            } else {
                strategies.add(ServiceBeanFactory.getBean(HistoryTitleStrategy.class));
                strategies.add(ServiceBeanFactory.getBean(LowScoreStrategy.class));
            }
        }
        return strategies;
    }

    private List<FilterStrategy> getPreFilterStrategies(FilterParam param) {
        List<FilterStrategy> strategies = new ArrayList<>();
        strategies.add(ServiceBeanFactory.getBean(BadStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(SensitiveStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(DeDuplicationStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(KeywordStrategy.class));
        strategies.add(ServiceBeanFactory.getBean(ArticlePromotionStrategy.class));
        if (param.getScene().equals(FWH_COLD_START)) {
            strategies.add(ServiceBeanFactory.getBean(HistoryTitleForFwhColdStartStrategy.class));
        } else {
            if (Objects.equals(param.getStrategy(), RankStrategyEnum.INFINITE_STRATEGY.getStrategy())) {
                strategies.add(ServiceBeanFactory.getBean(InfiniteHisTitleStrategy.class));
            } else {
                strategies.add(ServiceBeanFactory.getBean(HistoryTitleStrategy.class));
            }
        }
        return strategies;
    }

    private List<FilterStrategy> getUnPreFilterStrategies(FilterParam param) {
        List<FilterStrategy> strategies = new ArrayList<>();
        strategies.add(ServiceBeanFactory.getBean(TodayPublishStrategy.class));
        if (!param.getScene().equals(FWH_COLD_START)
                && !Objects.equals(param.getStrategy(), RankStrategyEnum.INFINITE_STRATEGY.getStrategy())) {
            strategies.add(ServiceBeanFactory.getBean(LowScoreStrategy.class));
        }
        return strategies;
    }
}
